/** Support functions for Unicode implementation
*/ 

#include <stdio.h>
#include <stdlib.h>

typedef unsigned short unichar;
typedef unsigned short BOOL;


typedef struct {unichar from; unsigned char to;} _ucc_;

#include <iconv.h>
#include <errno.h>

/*
 * The whole of the GNUstep code stores UNICODE in internal byte order,
 * so we do the same. This should be UCS-2-INTERNAL for libiconv
 */
#define UNICODE_INT "UNICODEBIG"

#define UNICODE_ENC UNICODE_INT

static const char *unicode_enc = NULL;

typedef	unsigned char	unc;
static char *	defEnc = 0;
static int *_availableEncodings = 0;

struct _strenc_ {
  int	enc;		// Constant representing the encoding.
  const char		*ename;		// ASCII string representation of name.
  const char		*iconv;		/* Iconv name of encoding.  If this
					 * is the empty string, we cannot use
					 * iconv perform conversions to/from
					 * this encoding.
					 * NB. do not put a nul pointer in this
					 * field in the table, use "" instread.
					 */
  BOOL			eightBit;	/* Flag to say whether this encoding
					 * can be stored in a byte array ...
					 * ie whether the encoding consists
					 * entirely of single byte charcters
					 * and the first 128 are identical to
					 * the ASCII character set.
					 */
  BOOL			supported;	/* Is this supported?  Some encodings
					 * have builtin conversion to/from
					 * unicode, but for others we must
					 * check with iconv to see if it
					 * supports them on this platform.
					 */
};

/*
 * The str_encoding_table is a compact representation of all the string
 * encoding information we might need.  It gets modified at runtime.
 */
static struct _strenc_ str_encoding_table[] = {
  {1,"NSASCIIStringEncoding","ASCII",1,1},
  {2,"NSNEXTSTEPStringEncoding","NEXTSTEP",1,1},
  {3, "NSJapaneseEUCStringEncoding","EUC-JP",0,0},
  {4,"NSUTF8StringEncoding","UTF-8",0,0},
  {5,"NSISOLatin1StringEncoding","ISO-8859-1",1,1},
  {6,"NSSymbolStringEncoding","",0,0},
  {7,"NSNonLossyASCIIStringEncoding","",1,1},
  {8,"NSShiftJISStringEncoding","SHIFT-JIS",0,0},
  {9,"NSISOLatin2StringEncoding","ISO-8859-2",1,1},
  {10, "NSUnicodeStringEncoding","",0,1},
  {11,"NSWindowsCP1251StringEncoding","CP1251",0,0},
  {12,"NSWindowsCP1252StringEncoding","CP1252",0,0},
  {13,"NSWindowsCP1253StringEncoding","CP1253",0,0},
  {14,"NSWindowsCP1254StringEncoding","CP1254",0,0},
  {15,"NSWindowsCP1250StringEncoding","CP1250",0,0},
  {16,"NSISO2022JPStringEncoding","ISO-2022-JP",0,0},
  {17, "NSMacOSRomanStringEncoding","MACINTOSH",0,0},
  {18, "NSProprietaryStringEncoding","",0,0},

// GNUstep additions
  {20,"NSISOCyrillicStringEncoding","ISO-8859-5",0,1},
  {21, "NSKOI8RStringEncoding","KOI8-R",0,0},
  {22, "NSISOLatin3StringEncoding","ISO-8859-3",0,0},
  {23, "NSISOLatin4StringEncoding","ISO-8859-4",0,0},
  {24, "NSISOArabicStringEncoding","ISO-8859-6",0,0},
  {25, "NSISOGreekStringEncoding","ISO-8859-7",0,0},
  {26, "NSISOHebrewStringEncoding","ISO-8859-8",0,0},
  {27, "NSISOLatin5StringEncoding","ISO-8859-9",0,0},
  {28, "NSISOLatin6StringEncoding","ISO-8859-10",0,0},
  {29, "NSISOLatin7StringEncoding","ISO-8859-13",0,0},
  {30, "NSISOLatin8StringEncoding","ISO-8859-14",0,0},
  {31, "NSISOLatin9StringEncoding","ISO-8859-15",0,0},
  {32, "NSUTF7StringEncoding","",0,0},
  {33, "NSGB2312StringEncoding","EUC-CN",0,0},
  {34, "NSGSM0338StringEncoding","",0,1},
  {35, "NSBIG5StringEncoding","BIG5",0,0},

  {0,"Unknown encoding","",0,0}
};

static struct _strenc_	**encodingTable = 0;
static unsigned		encTableSize = 0;

int *GetAvailableEncodings()
{
  if (_availableEncodings == 0)
    {
      if (_availableEncodings == 0)
	{
	  int	*encodings;
	  unsigned		count;
	  unsigned		pos;
	  unsigned		i;

	  /*
	   * We want to store pointers to our string encoding info in a
	   * large table so we can do efficient lookup by encoding value.
	   */
#define	MAX_ENCODING	128
	  count = sizeof(str_encoding_table) / sizeof(struct _strenc_);

	  /*
	   * First determine the largest encoding value and create a
	   * large enough table of pointers.
	   */
	  encTableSize = 0;
	  for (i = 0; i < count; i++)
	    {
	      unsigned	tmp = str_encoding_table[i].enc;

	      if (tmp >= MAX_ENCODING)
		{
		  fprintf(stderr, "ERROR ... illegal NSStringEncoding "
		    "value in str_encoding_table. Ignored\n");
		} 
	      else if (tmp > encTableSize)
		{
		  encTableSize = tmp;
		}
	    }
	  encodingTable = malloc((encTableSize+1)*sizeof(struct _strenc_ *));
	  memset(encodingTable, 0, (encTableSize+1)*sizeof(struct _strenc_ *));

	  /*
	   * Now set up the pointers at the correct location in the table.
	   */
	  for (i = 0; i < count; i++)
	    {
	      unsigned	tmp = str_encoding_table[i].enc;

	      if (tmp < MAX_ENCODING)
		{
		  encodingTable[tmp] = &str_encoding_table[i];
		}
	    }

	  /*
	   * Now build up a list of supported encodings ... in the
	   * format needed to support [NSStirng+availableStringEncodings]
	   * Check to see what iconv support we have as we go along.
	   * This is also the palce where we determine the name we use
	   * for iconv to support unicode.
	   */
	  encodings = malloc(sizeof(int) * count);
	  pos = 0;
	  for (i = 0; i < count; i++)
	    {
	      int	enc = str_encoding_table[i].enc;

	      if (enc == 0 || enc >= MAX_ENCODING)
		{
		  continue;
		}
	      if (enc == 10)
		{
		  encodingTable[enc]->iconv = UNICODE_ENC;
		  encodingTable[enc]->supported = 1;
		}
	      if (encodingTable[enc]->supported == 0)
		{
		  if (encodingTable[enc]->iconv == 0)
		    {
		      continue;		// Not handled by iconv.
		    }
		  else
		    {
		      iconv_t	c;
		      printf("Checking %s encoding\n",
				  encodingTable[enc]->iconv);
		      c = iconv_open(UNICODE_ENC, encodingTable[enc]->iconv);
		      printf("open = %d\n", c);
		      if (c == (iconv_t)-1)
			{
			  continue;	// Can't convert to unicode
			}
		      iconv_close(c);
		      printf("Checking reverse\n");
		      c = iconv_open(encodingTable[enc]->iconv, UNICODE_ENC);
		      printf("open = %d\n", c);
		      if (c == (iconv_t)-1)
			{
			  continue;	// Can't convert from unicode
			}
		      iconv_close(c);
		      encodingTable[enc]->supported = 1;
		    }
		}
	      encodings[pos++] = enc;
	    }
	  encodings[pos] = 0;
	  _availableEncodings = encodings;
	}
    }
  return _availableEncodings;
}
